"""Forms for the file_validator app."""
import csv
from collections.abc import Sequence
from typing import ClassVar

from django import forms


class CSVUploadForm(forms.Form):
    """Form for uploading CSV files only."""

    file = forms.FileField(label="Select a CSV file")

    required_columns: ClassVar[list] = [
        "AccountNumber", "CBAccountNumber", "DaysDiscountValid", "DiscountValue",
        "DiscountPercentage", "DueDate", "GoodsValueInAccountCurrency",
        "PurControlValueInBaseCurrency", "DocumentToBaseCurrencyRate",
        "DocumentToAccountCurrencyRate", "PostedDate", "QueryCode",
        "TransactionReference", "SecondReference", "Source",
        "SYSTraderTranType", "TransactionDate", "UniqueReferenceNumber",
        "UserNumber", "TaxValue", "SYSTraderGenerationReasonType",
        "GoodsValueInBaseCurrency"
    ]

    repeating_columns: ClassVar[dict] = {
        "NominalAnalysis": [
            "NominalAnalysisTransactionValue", "NominalAnalysisNominalAccountNumber",
            "NominalAnalysisNominalCostCentre", "NominalAnalysisNominalDepartment",
            "NominalAnalysisNominalAnalysisNarrative", "NominalAnalysisTransactionAnalysisCode"
        ],
        "TaxAnalysis": [
            "TaxAnalysisTaxRate", "TaxAnalysisGoodsValueBeforeDiscount",
            "TaxAnalysisDiscountValue", "TaxAnalysisDiscountPercentage",
            "TaxAnalysisTaxOnGoodsValue"
        ]
    }

    def clean_file(self) -> str:
        """Validate the uploaded file format and contents."""
        file = self.cleaned_data["file"]

        if not file.name.endswith(".csv"):
            err_msg = "File must be in CSV format"
            raise forms.ValidationError(err_msg)

        try:
            csv_file = file.read().decode("utf-8").splitlines()
            reader = csv.DictReader(csv_file)
            fieldnames = reader.fieldnames if reader.fieldnames is not None else []

            missing_columns = [col for col in self.required_columns if col not in fieldnames]

            for section_name, column_list in self.repeating_columns.items():
                max_repeat = self.get_max_repeat(fieldnames, section_name)

                if max_repeat == 0:
                    missing_columns.extend([f"{base_col}/1" for base_col in column_list])
                else:
                    for repeat in range(1, max_repeat + 1):
                        missing_columns.extend(
                            [f"{base_col}/{repeat}" for base_col in column_list if
                             f"{base_col}/{repeat}" not in fieldnames]
                        )

            if missing_columns:
                err_msg = f"Missing required columns: {", ".join(missing_columns)}"
                raise forms.ValidationError(err_msg)

        except (UnicodeDecodeError, csv.Error) as e:
            err_msg = f"File could not be processed: {e!s}"
            raise forms.ValidationError(err_msg) from e

        return file

    @staticmethod
    def get_max_repeat(fieldnames: Sequence[str], section_prefix: str) -> int:
        """Identify the maximum number of repeats for a section."""
        max_repeat = 0
        for field in fieldnames:
            if field.startswith(section_prefix):
                try:
                    repeat_number = int(field.split("/")[-1])
                    max_repeat = max(max_repeat, repeat_number)
                except ValueError:
                    continue
        return max_repeat
